@@ -1,13 +1,12 @@
 //! Instruction types
 
 #![allow(clippy::too_many_arguments)]
-
 use {
     crate::{
         find_deposit_authority_program_address, find_stake_program_address,
         find_transient_stake_program_address, find_withdraw_authority_program_address,
         stake_program,
-        state::{Fee, StakePool, ValidatorList},
+        state::{Fee, FeeType, StakePool, ValidatorList},
         MAX_VALIDATORS_TO_UPDATE,
     },
     borsh::{BorshDeserialize, BorshSchema, BorshSerialize},
@@ -29,6 +28,17 @@ pub enum PreferredValidatorType {
     Withdraw,
 }
 
+/// Defines which deposit authority to update in the `SetDepositAuthority`
+/// instruction
+#[repr(C)]
+#[derive(Clone, Debug, PartialEq, BorshSerialize, BorshDeserialize, BorshSchema)]
+pub enum DepositType {
+    /// Sets the stake deposit authority
+    Stake,
+    /// Sets the SOL deposit authority
+    Sol,
+}
+
 /// Instructions supported by the StakePool program.
 #[repr(C)]
 #[derive(Clone, Debug, PartialEq, BorshSerialize, BorshDeserialize, BorshSchema)]
@@ -53,47 +63,38 @@ pub enum StakePoolInstruction {
         /// Fee assessed as percentage of perceived rewards
         #[allow(dead_code)] // but it's not
         fee: Fee,
+        /// Fee charged per withdrawal as percentage of withdrawal
+        #[allow(dead_code)] // but it's not
+        withdrawal_fee: Fee,
+        /// Fee charged per deposit as percentage of deposit
+        #[allow(dead_code)] // but it's not
+        deposit_fee: Fee,
+        /// Percentage [0-100] of deposit_fee that goes to referrer
+        #[allow(dead_code)] // but it's not
+        referral_fee: u8,
         /// Maximum expected number of validators
         #[allow(dead_code)] // but it's not
         max_validators: u32,
     },
 
-    ///   (Staker only) Creates new program account for accumulating stakes for
-    ///   a particular validator
-    ///
-    ///   0. `[]` Stake pool account this stake will belong to
-    ///   1. `[s]` Staker
-    ///   2. `[ws]` Funding account (must be a system account)
-    ///   3. `[w]` Stake account to be created
-    ///   4. `[]` Validator this stake account will vote for
-    ///   5. `[]` Rent sysvar
-    ///   6. `[]` Stake History sysvar
-    ///   7. `[]` Stake Config sysvar
-    ///   8. `[]` System program
-    ///   9. `[]` Stake program
-    CreateValidatorStakeAccount,
-
     ///   (Staker only) Adds stake account delegated to validator to the pool's
     ///   list of managed validators.
     ///
-    ///   The stake account must have the rent-exempt amount plus at least 1 SOL,
-    ///   and at most 1.001 SOL.
-    ///
-    ///   Once we delegate even 1 SOL, it will accrue rewards one epoch later,
-    ///   so we'll have more than 1 active SOL at this point.
-    ///   At 10% annualized rewards, 1 epoch of 2 days will accrue
-    ///   0.000547945 SOL, so we check that it is at least 1 SOL, and at most
-    ///   1.001 SOL.
+    ///   The stake account will have the rent-exempt amount plus 1 SOL.
     ///
     ///   0. `[w]` Stake pool
     ///   1. `[s]` Staker
-    ///   2. `[]` Stake pool withdraw authority
-    ///   3. `[w]` Validator stake list storage account
-    ///   4. `[w]` Stake account to add to the pool, its withdraw authority must
-    ///      be set to the staker
-    ///   5. `[]` Clock sysvar
-    ///   6. '[]' Sysvar stake history account
-    ///   7. `[]` Stake program
+    ///   2. `[ws]` Funding account (must be a system account)
+    ///   3. `[]` Stake pool withdraw authority
+    ///   4. `[w]` Validator stake list storage account
+    ///   5. `[w]` Stake account to add to the pool
+    ///   6. `[]` Validator this stake account will be delegated to
+    ///   7. `[]` Rent sysvar
+    ///   8. `[]` Clock sysvar
+    ///   9. '[]' Stake history sysvar
+    ///  10. '[]' Stake config sysvar
+    ///  11. `[]` System program
+    ///  12. `[]` Stake program
     AddValidatorToPool,
 
     ///   (Staker only) Removes validator from the pool
@@ -108,8 +109,9 @@ pub enum StakePoolInstruction {
     ///   4. `[w]` Validator stake list storage account
     ///   5. `[w]` Stake account to remove from the pool
     ///   6. `[]` Transient stake account, to check that that we're not trying to activate
-    ///   7. '[]' Sysvar clock
-    ///   8. `[]` Stake program id,
+    ///   7. `[w]` Destination stake account, to receive the minimum SOL from the validator stake account. Must be
+    ///   8. `[]` Sysvar clock
+    ///   9. `[]` Stake program id,
     RemoveValidatorFromPool,
 
     /// (Staker only) Decrease active stake on a validator, eventually moving it to the reserve
@@ -136,8 +138,14 @@ pub enum StakePoolInstruction {
     ///  7. `[]` Rent sysvar
     ///  8. `[]` System program
     ///  9. `[]` Stake program
-    ///  userdata: amount of lamports to split into the transient stake account
-    DecreaseValidatorStake(u64),
+    DecreaseValidatorStake {
+        /// amount of lamports to split into the transient stake account
+        #[allow(dead_code)] // but it's not
+        lamports: u64,
+        /// seed used to create transient stake account
+        #[allow(dead_code)] // but it's not
+        transient_stake_seed: u64,
+    },
 
     /// (Staker only) Increase stake on a validator from the reserve account
     ///
@@ -167,7 +175,14 @@ pub enum StakePoolInstruction {
     ///  `lamports + stake_rent_exemption`
     ///  The rent-exemption of the stake account is withdrawn back to the reserve
     ///  after it is merged.
-    IncreaseValidatorStake(u64),
+    IncreaseValidatorStake {
+        /// amount of lamports to increase on the given validator
+        #[allow(dead_code)] // but it's not
+        lamports: u64,
+        /// seed used to create transient stake account
+        #[allow(dead_code)] // but it's not
+        transient_stake_seed: u64,
+    },
 
     /// (Staker only) Set the preferred deposit or withdraw stake account for the
     /// stake pool
@@ -242,18 +257,20 @@ pub enum StakePoolInstruction {
     ///
     ///   0. `[w]` Stake pool
     ///   1. `[w]` Validator stake list storage account
-    ///   2. `[]` Stake pool deposit authority
+    ///   2. `[s]/[]` Stake pool deposit authority
     ///   3. `[]` Stake pool withdraw authority
     ///   4. `[w]` Stake account to join the pool (withdraw authority for the stake account should be first set to the stake pool deposit authority)
     ///   5. `[w]` Validator stake account for the stake account to be merged with
     ///   6. `[w]` Reserve stake account, to withdraw rent exempt reserve
     ///   7. `[w]` User account to receive pool tokens
-    ///   8. `[w]` Pool token mint account
-    ///   9. '[]' Sysvar clock account
-    ///   10. '[]' Sysvar stake history account
-    ///   11. `[]` Pool token program id,
-    ///   12. `[]` Stake program id,
-    Deposit,
+    ///   8. `[w]` Account to receive pool fee tokens
+    ///   9. `[w]` Account to receive a portion of pool fee tokens as referral fees
+    ///   10. `[w]` Pool token mint account
+    ///   11. '[]' Sysvar clock account
+    ///   12. '[]' Sysvar stake history account
+    ///   13. `[]` Pool token program id,
+    ///   14. `[]` Stake program id,
+    DepositStake,
 
     ///   Withdraw the token from the pool at the current ratio.
     ///
@@ -273,12 +290,13 @@ pub enum StakePoolInstruction {
     ///   5. `[]` User account to set as a new withdraw authority
     ///   6. `[s]` User transfer authority, for pool token account
     ///   7. `[w]` User account with pool tokens to burn from
-    ///   8. `[w]` Pool token mint account
-    ///   9. `[]` Sysvar clock account (required)
-    ///  10. `[]` Pool token program id
-    ///  11. `[]` Stake program id,
+    ///   8. `[w]` Account to receive pool fee tokens
+    ///   9. `[w]` Pool token mint account
+    ///  10. `[]` Sysvar clock account (required)
+    ///  11. `[]` Pool token program id
+    ///  12. `[]` Stake program id,
     ///  userdata: amount of pool tokens to withdraw
-    Withdraw(u64),
+    WithdrawStake(u64),
 
     ///  (Manager only) Update manager
     ///
@@ -294,9 +312,9 @@ pub enum StakePoolInstruction {
     ///  1. `[s]` Manager
     ///  2. `[]` Sysvar clock
     SetFee {
-        /// Fee assessed as percentage of perceived rewards
+        /// Type of fee to update and value to update it to
         #[allow(dead_code)] // but it's not
-        fee: Fee,
+        fee: FeeType,
     },
 
     ///  (Manager or staker only) Update staker
@@ -305,6 +323,30 @@ pub enum StakePoolInstruction {
     ///  1. `[s]` Manager or current staker
     ///  2. '[]` New staker pubkey
     SetStaker,
+
+    ///   Deposit SOL directly into the pool's reserve account. The output is a "pool" token
+    ///   representing ownership into the pool. Inputs are converted to the current ratio.
+    ///
+    ///   0. `[w]` Stake pool
+    ///   1. `[s]/[]` Stake pool sol deposit authority.
+    ///   2. `[]` Stake pool withdraw authority
+    ///   3. `[w]` Reserve stake account, to withdraw rent exempt reserve
+    ///   4. `[s]` Account providing the lamports to be deposited into the pool
+    ///   5. `[w]` User account to receive pool tokens
+    ///   6. `[w]` Account to receive pool fee tokens
+    ///   7. `[w]` Account to receive a portion of pool fee tokens as referral fees
+    ///   8. `[w]` Pool token mint account
+    ///   9. '[]' Sysvar clock account
+    ///   10 `[]` System program account
+    ///   11. `[]` Pool token program id,
+    DepositSol(u64),
+
+    ///  (Manager only) Update SOL deposit authority
+    ///
+    ///  0. `[w]` StakePool
+    ///  1. `[s]` Manager
+    ///  2. '[]` New sol_deposit_authority pubkey or none
+    SetDepositAuthority(DepositType),
 }
 
 /// Creates an 'initialize' instruction.
@@ -320,10 +362,16 @@ pub fn initialize(
     token_program_id: &Pubkey,
     deposit_authority: Option<Pubkey>,
     fee: Fee,
+    withdrawal_fee: Fee,
+    deposit_fee: Fee,
+    referral_fee: u8,
     max_validators: u32,
 ) -> Instruction {
     let init_data = StakePoolInstruction::Initialize {
         fee,
+        withdrawal_fee,
+        deposit_fee,
+        referral_fee,
         max_validators,
     };
     let data = init_data.try_to_vec().unwrap();
@@ -349,20 +397,24 @@ pub fn initialize(
     }
 }
 
-/// Creates `CreateValidatorStakeAccount` instruction (create new stake account for the validator)
-pub fn create_validator_stake_account(
+/// Creates `AddValidatorToPool` instruction (add new validator stake account to the pool)
+pub fn add_validator_to_pool(
     program_id: &Pubkey,
     stake_pool: &Pubkey,
     staker: &Pubkey,
     funder: &Pubkey,
-    stake_account: &Pubkey,
+    stake_pool_withdraw: &Pubkey,
+    validator_list: &Pubkey,
+    stake: &Pubkey,
     validator: &Pubkey,
 ) -> Instruction {
     let accounts = vec![
-        AccountMeta::new_readonly(*stake_pool, false),
+        AccountMeta::new(*stake_pool, false),
         AccountMeta::new_readonly(*staker, true),
         AccountMeta::new(*funder, true),
-        AccountMeta::new(*stake_account, false),
+        AccountMeta::new_readonly(*stake_pool_withdraw, false),
+        AccountMeta::new(*validator_list, false),
+        AccountMeta::new(*stake, false),
         AccountMeta::new_readonly(*validator, false),
         AccountMeta::new_readonly(sysvar::rent::id(), false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
@@ -371,34 +423,6 @@ pub fn create_validator_stake_account(
         AccountMeta::new_readonly(system_program::id(), false),
         AccountMeta::new_readonly(stake_program::id(), false),
     ];
-    Instruction {
-        program_id: *program_id,
-        accounts,
-        data: StakePoolInstruction::CreateValidatorStakeAccount
-            .try_to_vec()
-            .unwrap(),
-    }
-}
-
-/// Creates `AddValidatorToPool` instruction (add new validator stake account to the pool)
-pub fn add_validator_to_pool(
-    program_id: &Pubkey,
-    stake_pool: &Pubkey,
-    staker: &Pubkey,
-    stake_pool_withdraw: &Pubkey,
-    validator_list: &Pubkey,
-    stake_account: &Pubkey,
-) -> Instruction {
-    let accounts = vec![
-        AccountMeta::new(*stake_pool, false),
-        AccountMeta::new_readonly(*staker, true),
-        AccountMeta::new_readonly(*stake_pool_withdraw, false),
-        AccountMeta::new(*validator_list, false),
-        AccountMeta::new(*stake_account, false),
-        AccountMeta::new_readonly(sysvar::clock::id(), false),
-        AccountMeta::new_readonly(sysvar::stake_history::id(), false),
-        AccountMeta::new_readonly(stake_program::id(), false),
-    ];
     Instruction {
         program_id: *program_id,
         accounts,
@@ -418,6 +442,7 @@ pub fn remove_validator_from_pool(
     validator_list: &Pubkey,
     stake_account: &Pubkey,
     transient_stake_account: &Pubkey,
+    destination_stake_account: &Pubkey,
 ) -> Instruction {
     let accounts = vec![
         AccountMeta::new(*stake_pool, false),
@@ -427,6 +452,7 @@ pub fn remove_validator_from_pool(
         AccountMeta::new(*validator_list, false),
         AccountMeta::new(*stake_account, false),
         AccountMeta::new_readonly(*transient_stake_account, false),
+        AccountMeta::new(*destination_stake_account, false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
         AccountMeta::new_readonly(stake_program::id(), false),
     ];
@@ -450,6 +476,7 @@ pub fn decrease_validator_stake(
     validator_stake: &Pubkey,
     transient_stake: &Pubkey,
     lamports: u64,
+    transient_stake_seed: u64,
 ) -> Instruction {
     let accounts = vec![
         AccountMeta::new_readonly(*stake_pool, false),
@@ -466,9 +493,12 @@ pub fn decrease_validator_stake(
     Instruction {
         program_id: *program_id,
         accounts,
-        data: StakePoolInstruction::DecreaseValidatorStake(lamports)
-            .try_to_vec()
-            .unwrap(),
+        data: StakePoolInstruction::DecreaseValidatorStake {
+            lamports,
+            transient_stake_seed,
+        }
+        .try_to_vec()
+        .unwrap(),
     }
 }
 
@@ -484,6 +514,7 @@ pub fn increase_validator_stake(
     transient_stake: &Pubkey,
     validator: &Pubkey,
     lamports: u64,
+    transient_stake_seed: u64,
 ) -> Instruction {
     let accounts = vec![
         AccountMeta::new_readonly(*stake_pool, false),
@@ -503,9 +534,12 @@ pub fn increase_validator_stake(
     Instruction {
         program_id: *program_id,
         accounts,
-        data: StakePoolInstruction::IncreaseValidatorStake(lamports)
-            .try_to_vec()
-            .unwrap(),
+        data: StakePoolInstruction::IncreaseValidatorStake {
+            lamports,
+            transient_stake_seed,
+        }
+        .try_to_vec()
+        .unwrap(),
     }
 }
 
@@ -534,32 +568,13 @@ pub fn set_preferred_validator(
     }
 }
 
-/// Creates `CreateValidatorStakeAccount` instruction with a vote account
-pub fn create_validator_stake_account_with_vote(
-    program_id: &Pubkey,
-    stake_pool_address: &Pubkey,
-    staker: &Pubkey,
-    funder: &Pubkey,
-    vote_account_address: &Pubkey,
-) -> Instruction {
-    let (stake_account, _) =
-        find_stake_program_address(program_id, vote_account_address, stake_pool_address);
-    create_validator_stake_account(
-        program_id,
-        stake_pool_address,
-        staker,
-        funder,
-        &stake_account,
-        vote_account_address,
-    )
-}
-
 /// Create an `AddValidatorToPool` instruction given an existing stake pool and
 /// vote account
 pub fn add_validator_to_pool_with_vote(
     program_id: &Pubkey,
     stake_pool: &StakePool,
     stake_pool_address: &Pubkey,
+    funder: &Pubkey,
     vote_account_address: &Pubkey,
 ) -> Instruction {
     let pool_withdraw_authority =
@@ -570,9 +585,11 @@ pub fn add_validator_to_pool_with_vote(
         program_id,
         stake_pool_address,
         &stake_pool.staker,
+        funder,
         &pool_withdraw_authority,
         &stake_pool.validator_list,
         &stake_account_address,
+        vote_account_address,
     )
 }
 
@@ -584,13 +601,19 @@ pub fn remove_validator_from_pool_with_vote(
     stake_pool_address: &Pubkey,
     vote_account_address: &Pubkey,
     new_stake_account_authority: &Pubkey,
+    transient_stake_seed: u64,
+    destination_stake_address: &Pubkey,
 ) -> Instruction {
     let pool_withdraw_authority =
         find_withdraw_authority_program_address(program_id, stake_pool_address).0;
     let (stake_account_address, _) =
         find_stake_program_address(program_id, vote_account_address, stake_pool_address);
-    let (transient_stake_account, _) =
-        find_transient_stake_program_address(program_id, vote_account_address, stake_pool_address);
+    let (transient_stake_account, _) = find_transient_stake_program_address(
+        program_id,
+        vote_account_address,
+        stake_pool_address,
+        transient_stake_seed,
+    );
     remove_validator_from_pool(
         program_id,
         stake_pool_address,
@@ -600,6 +623,7 @@ pub fn remove_validator_from_pool_with_vote(
         &stake_pool.validator_list,
         &stake_account_address,
         &transient_stake_account,
+        destination_stake_address,
     )
 }
 
@@ -611,11 +635,16 @@ pub fn increase_validator_stake_with_vote(
     stake_pool_address: &Pubkey,
     vote_account_address: &Pubkey,
     lamports: u64,
+    transient_stake_seed: u64,
 ) -> Instruction {
     let pool_withdraw_authority =
         find_withdraw_authority_program_address(program_id, stake_pool_address).0;
-    let (transient_stake_address, _) =
-        find_transient_stake_program_address(program_id, vote_account_address, stake_pool_address);
+    let (transient_stake_address, _) = find_transient_stake_program_address(
+        program_id,
+        vote_account_address,
+        stake_pool_address,
+        transient_stake_seed,
+    );
 
     increase_validator_stake(
         program_id,
@@ -627,6 +656,7 @@ pub fn increase_validator_stake_with_vote(
         &transient_stake_address,
         vote_account_address,
         lamports,
+        transient_stake_seed,
     )
 }
 
@@ -638,13 +668,18 @@ pub fn decrease_validator_stake_with_vote(
     stake_pool_address: &Pubkey,
     vote_account_address: &Pubkey,
     lamports: u64,
+    transient_stake_seed: u64,
 ) -> Instruction {
     let pool_withdraw_authority =
         find_withdraw_authority_program_address(program_id, stake_pool_address).0;
     let (validator_stake_address, _) =
         find_stake_program_address(program_id, vote_account_address, stake_pool_address);
-    let (transient_stake_address, _) =
-        find_transient_stake_program_address(program_id, vote_account_address, stake_pool_address);
+    let (transient_stake_address, _) = find_transient_stake_program_address(
+        program_id,
+        vote_account_address,
+        stake_pool_address,
+        transient_stake_seed,
+    );
     decrease_validator_stake(
         program_id,
         stake_pool_address,
@@ -654,6 +689,7 @@ pub fn decrease_validator_stake_with_vote(
         &validator_stake_address,
         &transient_stake_address,
         lamports,
+        transient_stake_seed,
     )
 }
 
@@ -662,8 +698,9 @@ pub fn update_validator_list_balance(
     program_id: &Pubkey,
     stake_pool: &Pubkey,
     stake_pool_withdraw_authority: &Pubkey,
-    validator_list: &Pubkey,
+    validator_list_address: &Pubkey,
     reserve_stake: &Pubkey,
+    validator_list: &ValidatorList,
     validator_vote_accounts: &[Pubkey],
     start_index: u32,
     no_merge: bool,
@@ -671,7 +708,7 @@ pub fn update_validator_list_balance(
     let mut accounts = vec![
         AccountMeta::new_readonly(*stake_pool, false),
         AccountMeta::new_readonly(*stake_pool_withdraw_authority, false),
-        AccountMeta::new(*validator_list, false),
+        AccountMeta::new(*validator_list_address, false),
         AccountMeta::new(*reserve_stake, false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
         AccountMeta::new_readonly(sysvar::stake_history::id(), false),
@@ -681,17 +718,23 @@ pub fn update_validator_list_balance(
         &mut validator_vote_accounts
             .iter()
             .flat_map(|vote_account_address| {
-                let (validator_stake_account, _) =
-                    find_stake_program_address(program_id, vote_account_address, stake_pool);
-                let (transient_stake_account, _) = find_transient_stake_program_address(
-                    program_id,
-                    vote_account_address,
-                    stake_pool,
-                );
-                vec![
-                    AccountMeta::new(validator_stake_account, false),
-                    AccountMeta::new(transient_stake_account, false),
-                ]
+                let validator_stake_info = validator_list.find(vote_account_address);
+                if let Some(validator_stake_info) = validator_stake_info {
+                    let (validator_stake_account, _) =
+                        find_stake_program_address(program_id, vote_account_address, stake_pool);
+                    let (transient_stake_account, _) = find_transient_stake_program_address(
+                        program_id,
+                        vote_account_address,
+                        stake_pool,
+                        validator_stake_info.transient_seed_suffix_start,
+                    );
+                    vec![
+                        AccountMeta::new(validator_stake_account, false),
+                        AccountMeta::new(transient_stake_account, false),
+                    ]
+                } else {
+                    vec![]
+                }
             })
             .collect::<Vec<AccountMeta>>(),
     );
@@ -716,6 +759,7 @@ pub fn update_stake_pool_balance(
     reserve_stake: &Pubkey,
     manager_fee_account: &Pubkey,
     stake_pool_mint: &Pubkey,
+    token_program_id: &Pubkey,
 ) -> Instruction {
     let accounts = vec![
         AccountMeta::new(*stake_pool, false),
@@ -725,7 +769,7 @@ pub fn update_stake_pool_balance(
         AccountMeta::new(*manager_fee_account, false),
         AccountMeta::new(*stake_pool_mint, false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
-        AccountMeta::new_readonly(spl_token::id(), false),
+        AccountMeta::new_readonly(*token_program_id, false),
     ];
     Instruction {
         program_id: *program_id,
@@ -782,6 +826,7 @@ pub fn update_stake_pool(
             &withdraw_authority,
             &stake_pool.validator_list,
             &stake_pool.reserve_stake,
+            validator_list,
             accounts_chunk,
             start_index,
             no_merge,
@@ -798,6 +843,7 @@ pub fn update_stake_pool(
             &stake_pool.reserve_stake,
             &stake_pool.manager_fee_account,
             &stake_pool.pool_mint,
+            &stake_pool.token_program_id,
         ),
         cleanup_removed_validator_entries(
             program_id,
@@ -810,7 +856,7 @@ pub fn update_stake_pool(
 
 /// Creates instructions required to deposit into a stake pool, given a stake
 /// account owned by the user.
-pub fn deposit(
+pub fn deposit_stake(
     program_id: &Pubkey,
     stake_pool: &Pubkey,
     validator_list_storage: &Pubkey,
@@ -820,6 +866,8 @@ pub fn deposit(
     validator_stake_account: &Pubkey,
     reserve_stake_account: &Pubkey,
     pool_tokens_to: &Pubkey,
+    manager_fee_account: &Pubkey,
+    referrer_pool_tokens_account: &Pubkey,
     pool_mint: &Pubkey,
     token_program_id: &Pubkey,
 ) -> Vec<Instruction> {
@@ -834,6 +882,8 @@ pub fn deposit(
         AccountMeta::new(*validator_stake_account, false),
         AccountMeta::new(*reserve_stake_account, false),
         AccountMeta::new(*pool_tokens_to, false),
+        AccountMeta::new(*manager_fee_account, false),
+        AccountMeta::new(*referrer_pool_tokens_account, false),
         AccountMeta::new(*pool_mint, false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
         AccountMeta::new_readonly(sysvar::stake_history::id(), false),
@@ -856,7 +906,7 @@ pub fn deposit(
         Instruction {
             program_id: *program_id,
             accounts,
-            data: StakePoolInstruction::Deposit.try_to_vec().unwrap(),
+            data: StakePoolInstruction::DepositStake.try_to_vec().unwrap(),
         },
     ]
 }
@@ -864,7 +914,7 @@ pub fn deposit(
 /// Creates instructions required to deposit into a stake pool, given a stake
 /// account owned by the user. The difference with `deposit()` is that a deposit
 /// authority must sign this instruction, which is required for private pools.
-pub fn deposit_with_authority(
+pub fn deposit_stake_with_authority(
     program_id: &Pubkey,
     stake_pool: &Pubkey,
     validator_list_storage: &Pubkey,
@@ -875,6 +925,8 @@ pub fn deposit_with_authority(
     validator_stake_account: &Pubkey,
     reserve_stake_account: &Pubkey,
     pool_tokens_to: &Pubkey,
+    manager_fee_account: &Pubkey,
+    referrer_pool_tokens_account: &Pubkey,
     pool_mint: &Pubkey,
     token_program_id: &Pubkey,
 ) -> Vec<Instruction> {
@@ -887,6 +939,8 @@ pub fn deposit_with_authority(
         AccountMeta::new(*validator_stake_account, false),
         AccountMeta::new(*reserve_stake_account, false),
         AccountMeta::new(*pool_tokens_to, false),
+        AccountMeta::new(*manager_fee_account, false),
+        AccountMeta::new(*referrer_pool_tokens_account, false),
         AccountMeta::new(*pool_mint, false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
         AccountMeta::new_readonly(sysvar::stake_history::id(), false),
@@ -909,13 +963,91 @@ pub fn deposit_with_authority(
         Instruction {
             program_id: *program_id,
             accounts,
-            data: StakePoolInstruction::Deposit.try_to_vec().unwrap(),
+            data: StakePoolInstruction::DepositStake.try_to_vec().unwrap(),
         },
     ]
 }
 
-/// Creates a 'withdraw' instruction.
-pub fn withdraw(
+/// Creates instructions required to deposit SOL directly into a stake pool.
+pub fn deposit_sol(
+    program_id: &Pubkey,
+    stake_pool: &Pubkey,
+    stake_pool_withdraw_authority: &Pubkey,
+    reserve_stake_account: &Pubkey,
+    lamports_from: &Pubkey,
+    pool_tokens_to: &Pubkey,
+    manager_fee_account: &Pubkey,
+    referrer_pool_tokens_account: &Pubkey,
+    pool_mint: &Pubkey,
+    token_program_id: &Pubkey,
+    amount: u64,
+) -> Vec<Instruction> {
+    let accounts = vec![
+        AccountMeta::new(*stake_pool, false),
+        AccountMeta::new_readonly(*stake_pool_withdraw_authority, false),
+        AccountMeta::new(*reserve_stake_account, false),
+        AccountMeta::new(*lamports_from, true),
+        AccountMeta::new(*pool_tokens_to, false),
+        AccountMeta::new(*manager_fee_account, false),
+        AccountMeta::new(*referrer_pool_tokens_account, false),
+        AccountMeta::new(*pool_mint, false),
+        AccountMeta::new_readonly(sysvar::clock::id(), false),
+        AccountMeta::new_readonly(system_program::id(), false),
+        AccountMeta::new_readonly(*token_program_id, false),
+    ];
+    vec![Instruction {
+        program_id: *program_id,
+        accounts,
+        data: StakePoolInstruction::DepositSol(amount)
+            .try_to_vec()
+            .unwrap(),
+    }]
+}
+
+/// Creates instructions required to deposit SOL directly into a stake pool.
+/// The difference with `deposit_sol()` is that a deposit
+/// authority must sign this instruction, which is required for private pools.
+/// `require_deposit_authority` should be false only if
+/// `sol_deposit_authority == None`
+pub fn deposit_sol_with_authority(
+    program_id: &Pubkey,
+    stake_pool: &Pubkey,
+    sol_deposit_authority: &Pubkey,
+    stake_pool_withdraw_authority: &Pubkey,
+    reserve_stake_account: &Pubkey,
+    lamports_from: &Pubkey,
+    pool_tokens_to: &Pubkey,
+    manager_fee_account: &Pubkey,
+    referrer_pool_tokens_account: &Pubkey,
+    pool_mint: &Pubkey,
+    token_program_id: &Pubkey,
+    amount: u64,
+) -> Vec<Instruction> {
+    let accounts = vec![
+        AccountMeta::new(*stake_pool, false),
+        AccountMeta::new_readonly(*stake_pool_withdraw_authority, false),
+        AccountMeta::new(*reserve_stake_account, false),
+        AccountMeta::new(*lamports_from, true),
+        AccountMeta::new(*pool_tokens_to, false),
+        AccountMeta::new(*manager_fee_account, false),
+        AccountMeta::new(*referrer_pool_tokens_account, false),
+        AccountMeta::new(*pool_mint, false),
+        AccountMeta::new_readonly(sysvar::clock::id(), false),
+        AccountMeta::new_readonly(system_program::id(), false),
+        AccountMeta::new_readonly(*token_program_id, false),
+        AccountMeta::new_readonly(*sol_deposit_authority, true),
+    ];
+    vec![Instruction {
+        program_id: *program_id,
+        accounts,
+        data: StakePoolInstruction::DepositSol(amount)
+            .try_to_vec()
+            .unwrap(),
+    }]
+}
+
+/// Creates a 'WithdrawStake' instruction.
+pub fn withdraw_stake(
     program_id: &Pubkey,
     stake_pool: &Pubkey,
     validator_list_storage: &Pubkey,
@@ -925,6 +1057,7 @@ pub fn withdraw(
     user_stake_authority: &Pubkey,
     user_transfer_authority: &Pubkey,
     user_pool_token_account: &Pubkey,
+    manager_fee_account: &Pubkey,
     pool_mint: &Pubkey,
     token_program_id: &Pubkey,
     amount: u64,
@@ -938,6 +1071,7 @@ pub fn withdraw(
         AccountMeta::new_readonly(*user_stake_authority, false),
         AccountMeta::new_readonly(*user_transfer_authority, true),
         AccountMeta::new(*user_pool_token_account, false),
+        AccountMeta::new(*manager_fee_account, false),
         AccountMeta::new(*pool_mint, false),
         AccountMeta::new_readonly(sysvar::clock::id(), false),
         AccountMeta::new_readonly(*token_program_id, false),
@@ -946,7 +1080,9 @@ pub fn withdraw(
     Instruction {
         program_id: *program_id,
         accounts,
-        data: StakePoolInstruction::Withdraw(amount).try_to_vec().unwrap(),
+        data: StakePoolInstruction::WithdrawStake(amount)
+            .try_to_vec()
+            .unwrap(),
     }
 }
 
@@ -961,7 +1097,7 @@ pub fn set_manager(
     let accounts = vec![
         AccountMeta::new(*stake_pool, false),
         AccountMeta::new_readonly(*manager, true),
-        AccountMeta::new_readonly(*new_manager, false),
+        AccountMeta::new_readonly(*new_manager, true),
         AccountMeta::new_readonly(*new_fee_receiver, false),
     ];
     Instruction {
@@ -976,7 +1112,7 @@ pub fn set_fee(
     program_id: &Pubkey,
     stake_pool: &Pubkey,
     manager: &Pubkey,
-    fee: Fee,
+    fee: FeeType,
 ) -> Instruction {
     let accounts = vec![
         AccountMeta::new(*stake_pool, false),
@@ -1008,3 +1144,27 @@ pub fn set_staker(
         data: StakePoolInstruction::SetStaker.try_to_vec().unwrap(),
     }
 }
+
+/// Creates a 'set deposit authority' instruction.
+pub fn set_deposit_authority(
+    program_id: &Pubkey,
+    stake_pool: &Pubkey,
+    manager: &Pubkey,
+    new_sol_deposit_authority: Option<&Pubkey>,
+    deposit_type: DepositType,
+) -> Instruction {
+    let mut accounts = vec![
+        AccountMeta::new(*stake_pool, false),
+        AccountMeta::new_readonly(*manager, true),
+    ];
+    if let Some(auth) = new_sol_deposit_authority {
+        accounts.push(AccountMeta::new_readonly(*auth, false))
+    }
+    Instruction {
+        program_id: *program_id,
+        accounts,
+        data: StakePoolInstruction::SetDepositAuthority(deposit_type)
+            .try_to_vec()
+            .unwrap(),
+    }
+}
