@@ -21,12 +21,15 @@ use {
     },
     spl_stake_pool::{
         find_stake_program_address, find_transient_stake_program_address, id, instruction,
-        processor, stake_program, state,
+        processor, stake_program,
+        state::{self, FeeType, ValidatorList},
+        MINIMUM_ACTIVE_STAKE,
     },
 };
 
 pub const TEST_STAKE_AMOUNT: u64 = 1_500_000_000;
 pub const MAX_TEST_VALIDATORS: u32 = 10_000;
+pub const DEFAULT_TRANSIENT_STAKE_SEED: u64 = 42;
 
 pub fn program_test() -> ProgramTest {
     ProgramTest::new(
@@ -99,6 +102,32 @@ pub async fn transfer(
     banks_client.process_transaction(transaction).await.unwrap();
 }
 
+pub async fn transfer_spl_tokens(
+    banks_client: &mut BanksClient,
+    payer: &Keypair,
+    recent_blockhash: &Hash,
+    source: &Pubkey,
+    destination: &Pubkey,
+    authority: &Keypair,
+    amount: u64,
+) {
+    let transaction = Transaction::new_signed_with_payer(
+        &[spl_token::instruction::transfer(
+            &spl_token::id(),
+            source,
+            destination,
+            &authority.pubkey(),
+            &[],
+            amount,
+        )
+        .unwrap()],
+        Some(&payer.pubkey()),
+        &[payer, authority],
+        *recent_blockhash,
+    );
+    banks_client.process_transaction(transaction).await.unwrap();
+}
+
 pub async fn create_token_account(
     banks_client: &mut BanksClient,
     payer: &Keypair,
@@ -134,6 +163,54 @@ pub async fn create_token_account(
     Ok(())
 }
 
+pub async fn close_token_account(
+    banks_client: &mut BanksClient,
+    payer: &Keypair,
+    recent_blockhash: &Hash,
+    account: &Pubkey,
+    lamports_destination: &Pubkey,
+    manager: &Keypair,
+) -> Result<(), TransportError> {
+    let mut transaction = Transaction::new_with_payer(
+        &[spl_token::instruction::close_account(
+            &spl_token::id(),
+            &account,
+            &lamports_destination,
+            &manager.pubkey(),
+            &[],
+        )
+        .unwrap()],
+        Some(&payer.pubkey()),
+    );
+    transaction.sign(&[payer, manager], *recent_blockhash);
+    banks_client.process_transaction(transaction).await?;
+    Ok(())
+}
+
+pub async fn freeze_token_account(
+    banks_client: &mut BanksClient,
+    payer: &Keypair,
+    recent_blockhash: &Hash,
+    account: &Pubkey,
+    pool_mint: &Pubkey,
+    manager: &Keypair,
+) -> Result<(), TransportError> {
+    let mut transaction = Transaction::new_with_payer(
+        &[spl_token::instruction::freeze_account(
+            &spl_token::id(),
+            &account,
+            pool_mint,
+            &manager.pubkey(),
+            &[],
+        )
+        .unwrap()],
+        Some(&payer.pubkey()),
+    );
+    transaction.sign(&[payer, manager], *recent_blockhash);
+    banks_client.process_transaction(transaction).await?;
+    Ok(())
+}
+
 pub async fn mint_tokens(
     banks_client: &mut BanksClient,
     payer: &Keypair,
@@ -161,6 +238,33 @@ pub async fn mint_tokens(
     Ok(())
 }
 
+pub async fn burn_tokens(
+    banks_client: &mut BanksClient,
+    payer: &Keypair,
+    recent_blockhash: &Hash,
+    mint: &Pubkey,
+    account: &Pubkey,
+    authority: &Keypair,
+    amount: u64,
+) -> Result<(), TransportError> {
+    let transaction = Transaction::new_signed_with_payer(
+        &[spl_token::instruction::burn(
+            &spl_token::id(),
+            account,
+            mint,
+            &authority.pubkey(),
+            &[],
+            amount,
+        )
+        .unwrap()],
+        Some(&payer.pubkey()),
+        &[payer, authority],
+        *recent_blockhash,
+    );
+    banks_client.process_transaction(transaction).await?;
+    Ok(())
+}
+
 pub async fn get_token_balance(banks_client: &mut BanksClient, token: &Pubkey) -> u64 {
     let token_account = banks_client.get_account(*token).await.unwrap().unwrap();
     let account_info: spl_token::state::Account =
@@ -213,8 +317,13 @@ pub async fn create_stake_pool(
     pool_token_account: &Pubkey,
     manager: &Keypair,
     staker: &Pubkey,
-    deposit_authority: &Option<Keypair>,
+    stake_deposit_authority: &Option<Keypair>,
     fee: &state::Fee,
+    withdrawal_fee: &state::Fee,
+    deposit_fee: &state::Fee,
+    referral_fee: u8,
+    sol_deposit_fee: &state::Fee,
+    sol_referral_fee: u8,
     max_validators: u32,
 ) -> Result<(), TransportError> {
     let rent = banks_client.get_rent().await.unwrap();
@@ -249,16 +358,31 @@ pub async fn create_stake_pool(
                 pool_mint,
                 pool_token_account,
                 &spl_token::id(),
-                deposit_authority.as_ref().map(|k| k.pubkey()),
+                stake_deposit_authority.as_ref().map(|k| k.pubkey()),
                 *fee,
+                *withdrawal_fee,
+                *deposit_fee,
+                referral_fee,
                 max_validators,
             ),
+            instruction::set_fee(
+                &id(),
+                &stake_pool.pubkey(),
+                &manager.pubkey(),
+                FeeType::SolDeposit(*sol_deposit_fee),
+            ),
+            instruction::set_fee(
+                &id(),
+                &stake_pool.pubkey(),
+                &manager.pubkey(),
+                FeeType::SolReferral(sol_referral_fee),
+            ),
         ],
         Some(&payer.pubkey()),
     );
     let mut signers = vec![payer, stake_pool, validator_list, manager];
-    if let Some(deposit_authority) = deposit_authority.as_ref() {
-        signers.push(deposit_authority);
+    if let Some(stake_deposit_authority) = stake_deposit_authority.as_ref() {
+        signers.push(stake_deposit_authority);
     }
     transaction.sign(&signers, *recent_blockhash);
     banks_client.process_transaction(transaction).await?;
@@ -358,31 +482,6 @@ pub async fn create_blank_stake_account(
     lamports
 }
 
-pub async fn create_validator_stake_account(
-    banks_client: &mut BanksClient,
-    payer: &Keypair,
-    recent_blockhash: &Hash,
-    stake_pool: &Pubkey,
-    staker: &Keypair,
-    stake_account: &Pubkey,
-    validator: &Pubkey,
-) {
-    let transaction = Transaction::new_signed_with_payer(
-        &[instruction::create_validator_stake_account(
-            &id(),
-            &stake_pool,
-            &staker.pubkey(),
-            &payer.pubkey(),
-            &stake_account,
-            &validator,
-        )],
-        Some(&payer.pubkey()),
-        &[payer, staker],
-        *recent_blockhash,
-    );
-    banks_client.process_transaction(transaction).await.unwrap();
-}
-
 pub async fn delegate_stake_account(
     banks_client: &mut BanksClient,
     payer: &Keypair,
@@ -425,57 +524,78 @@ pub async fn authorize_stake_account(
     banks_client.process_transaction(transaction).await.unwrap();
 }
 
+pub async fn create_unknown_validator_stake(
+    banks_client: &mut BanksClient,
+    payer: &Keypair,
+    recent_blockhash: &Hash,
+    stake_pool: &Pubkey,
+) -> ValidatorStakeAccount {
+    let mut unknown_stake = ValidatorStakeAccount::new(stake_pool, 222);
+    create_vote(
+        banks_client,
+        payer,
+        recent_blockhash,
+        &unknown_stake.validator,
+        &unknown_stake.vote,
+    )
+    .await;
+    let user = Keypair::new();
+    let fake_validator_stake = Keypair::new();
+    create_independent_stake_account(
+        banks_client,
+        payer,
+        recent_blockhash,
+        &fake_validator_stake,
+        &stake_program::Authorized {
+            staker: user.pubkey(),
+            withdrawer: user.pubkey(),
+        },
+        &stake_program::Lockup::default(),
+        MINIMUM_ACTIVE_STAKE,
+    )
+    .await;
+    delegate_stake_account(
+        banks_client,
+        payer,
+        recent_blockhash,
+        &fake_validator_stake.pubkey(),
+        &user,
+        &unknown_stake.vote.pubkey(),
+    )
+    .await;
+    unknown_stake.stake_account = fake_validator_stake.pubkey();
+    unknown_stake
+}
+
 pub struct ValidatorStakeAccount {
     pub stake_account: Pubkey,
     pub transient_stake_account: Pubkey,
+    pub transient_stake_seed: u64,
     pub vote: Keypair,
     pub validator: Keypair,
     pub stake_pool: Pubkey,
 }
 
 impl ValidatorStakeAccount {
-    pub fn new(stake_pool: &Pubkey) -> Self {
+    pub fn new(stake_pool: &Pubkey, transient_stake_seed: u64) -> Self {
         let validator = Keypair::new();
         let vote = Keypair::new();
         let (stake_account, _) = find_stake_program_address(&id(), &vote.pubkey(), stake_pool);
-        let (transient_stake_account, _) =
-            find_transient_stake_program_address(&id(), &vote.pubkey(), stake_pool);
+        let (transient_stake_account, _) = find_transient_stake_program_address(
+            &id(),
+            &vote.pubkey(),
+            stake_pool,
+            transient_stake_seed,
+        );
         ValidatorStakeAccount {
             stake_account,
             transient_stake_account,
+            transient_stake_seed,
             vote,
             validator,
             stake_pool: *stake_pool,
         }
     }
-
-    pub async fn create_and_delegate(
-        &self,
-        mut banks_client: &mut BanksClient,
-        payer: &Keypair,
-        recent_blockhash: &Hash,
-        staker: &Keypair,
-    ) {
-        create_vote(
-            &mut banks_client,
-            &payer,
-            &recent_blockhash,
-            &self.validator,
-            &self.vote,
-        )
-        .await;
-
-        create_validator_stake_account(
-            &mut banks_client,
-            &payer,
-            &recent_blockhash,
-            &self.stake_pool,
-            staker,
-            &self.stake_account,
-            &self.vote.pubkey(),
-        )
-        .await;
-    }
 }
 
 pub struct StakePoolAccounts {
@@ -487,9 +607,14 @@ pub struct StakePoolAccounts {
     pub manager: Keypair,
     pub staker: Keypair,
     pub withdraw_authority: Pubkey,
-    pub deposit_authority: Pubkey,
-    pub deposit_authority_keypair: Option<Keypair>,
+    pub stake_deposit_authority: Pubkey,
+    pub stake_deposit_authority_keypair: Option<Keypair>,
     pub fee: state::Fee,
+    pub withdrawal_fee: state::Fee,
+    pub deposit_fee: state::Fee,
+    pub referral_fee: u8,
+    pub sol_deposit_fee: state::Fee,
+    pub sol_referral_fee: u8,
     pub max_validators: u32,
 }
 
@@ -502,7 +627,7 @@ impl StakePoolAccounts {
             &[&stake_pool_address.to_bytes()[..32], b"withdraw"],
             &id(),
         );
-        let (deposit_authority, _) = Pubkey::find_program_address(
+        let (stake_deposit_authority, _) = Pubkey::find_program_address(
             &[&stake_pool_address.to_bytes()[..32], b"deposit"],
             &id(),
         );
@@ -521,20 +646,34 @@ impl StakePoolAccounts {
             manager,
             staker,
             withdraw_authority,
-            deposit_authority,
-            deposit_authority_keypair: None,
+            stake_deposit_authority,
+            stake_deposit_authority_keypair: None,
             fee: state::Fee {
                 numerator: 1,
                 denominator: 100,
             },
+            withdrawal_fee: state::Fee {
+                numerator: 3,
+                denominator: 1000,
+            },
+            deposit_fee: state::Fee {
+                numerator: 1,
+                denominator: 1000,
+            },
+            referral_fee: 25,
+            sol_deposit_fee: state::Fee {
+                numerator: 3,
+                denominator: 100,
+            },
+            sol_referral_fee: 50,
             max_validators: MAX_TEST_VALIDATORS,
         }
     }
 
-    pub fn new_with_deposit_authority(deposit_authority: Keypair) -> Self {
+    pub fn new_with_stake_deposit_authority(stake_deposit_authority: Keypair) -> Self {
         let mut stake_pool_accounts = Self::new();
-        stake_pool_accounts.deposit_authority = deposit_authority.pubkey();
-        stake_pool_accounts.deposit_authority_keypair = Some(deposit_authority);
+        stake_pool_accounts.stake_deposit_authority = stake_deposit_authority.pubkey();
+        stake_pool_accounts.stake_deposit_authority_keypair = Some(stake_deposit_authority);
         stake_pool_accounts
     }
 
@@ -542,6 +681,26 @@ impl StakePoolAccounts {
         amount * self.fee.numerator / self.fee.denominator
     }
 
+    pub fn calculate_withdrawal_fee(&self, pool_tokens: u64) -> u64 {
+        pool_tokens * self.withdrawal_fee.numerator / self.withdrawal_fee.denominator
+    }
+
+    pub fn calculate_deposit_fee(&self, pool_tokens: u64) -> u64 {
+        pool_tokens * self.deposit_fee.numerator / self.deposit_fee.denominator
+    }
+
+    pub fn calculate_referral_fee(&self, deposit_fee_collected: u64) -> u64 {
+        deposit_fee_collected * self.referral_fee as u64 / 100
+    }
+
+    pub fn calculate_sol_deposit_fee(&self, pool_tokens: u64) -> u64 {
+        pool_tokens * self.sol_deposit_fee.numerator / self.sol_deposit_fee.denominator
+    }
+
+    pub fn calculate_sol_referral_fee(&self, deposit_fee_collected: u64) -> u64 {
+        deposit_fee_collected * self.sol_referral_fee as u64 / 100
+    }
+
     pub async fn initialize_stake_pool(
         &self,
         mut banks_client: &mut BanksClient,
@@ -590,11 +749,17 @@ impl StakePoolAccounts {
             &self.pool_fee_account.pubkey(),
             &self.manager,
             &self.staker.pubkey(),
-            &self.deposit_authority_keypair,
+            &self.stake_deposit_authority_keypair,
             &self.fee,
+            &self.withdrawal_fee,
+            &self.deposit_fee,
+            self.referral_fee,
+            &self.sol_deposit_fee,
+            self.sol_referral_fee,
             self.max_validators,
         )
         .await?;
+
         Ok(())
     }
 
@@ -608,38 +773,118 @@ impl StakePoolAccounts {
         pool_account: &Pubkey,
         validator_stake_account: &Pubkey,
         current_staker: &Keypair,
+    ) -> Option<TransportError> {
+        self.deposit_stake_with_referral(
+            banks_client,
+            payer,
+            recent_blockhash,
+            stake,
+            pool_account,
+            validator_stake_account,
+            current_staker,
+            &self.pool_fee_account.pubkey(),
+        )
+        .await
+    }
+
+    #[allow(clippy::too_many_arguments)]
+    pub async fn deposit_stake_with_referral(
+        &self,
+        banks_client: &mut BanksClient,
+        payer: &Keypair,
+        recent_blockhash: &Hash,
+        stake: &Pubkey,
+        pool_account: &Pubkey,
+        validator_stake_account: &Pubkey,
+        current_staker: &Keypair,
+        referrer: &Pubkey,
     ) -> Option<TransportError> {
         let mut signers = vec![payer, current_staker];
-        let instructions = if let Some(deposit_authority) = self.deposit_authority_keypair.as_ref()
-        {
-            signers.push(deposit_authority);
-            instruction::deposit_with_authority(
+        let instructions =
+            if let Some(stake_deposit_authority) = self.stake_deposit_authority_keypair.as_ref() {
+                signers.push(stake_deposit_authority);
+                instruction::deposit_stake_with_authority(
+                    &id(),
+                    &self.stake_pool.pubkey(),
+                    &self.validator_list.pubkey(),
+                    &self.stake_deposit_authority,
+                    &self.withdraw_authority,
+                    stake,
+                    &current_staker.pubkey(),
+                    validator_stake_account,
+                    &self.reserve_stake.pubkey(),
+                    pool_account,
+                    &self.pool_fee_account.pubkey(),
+                    referrer,
+                    &self.pool_mint.pubkey(),
+                    &spl_token::id(),
+                )
+            } else {
+                instruction::deposit_stake(
+                    &id(),
+                    &self.stake_pool.pubkey(),
+                    &self.validator_list.pubkey(),
+                    &self.withdraw_authority,
+                    stake,
+                    &current_staker.pubkey(),
+                    validator_stake_account,
+                    &self.reserve_stake.pubkey(),
+                    pool_account,
+                    &self.pool_fee_account.pubkey(),
+                    &referrer,
+                    &self.pool_mint.pubkey(),
+                    &spl_token::id(),
+                )
+            };
+        let transaction = Transaction::new_signed_with_payer(
+            &instructions,
+            Some(&payer.pubkey()),
+            &signers,
+            *recent_blockhash,
+        );
+        banks_client.process_transaction(transaction).await.err()
+    }
+
+    #[allow(clippy::too_many_arguments)]
+    pub async fn deposit_sol(
+        &self,
+        banks_client: &mut BanksClient,
+        payer: &Keypair,
+        recent_blockhash: &Hash,
+        pool_account: &Pubkey,
+        amount: u64,
+        sol_deposit_authority: Option<&Keypair>,
+    ) -> Option<TransportError> {
+        let mut signers = vec![payer];
+        let instructions = if let Some(sol_deposit_authority) = sol_deposit_authority {
+            signers.push(sol_deposit_authority);
+            instruction::deposit_sol_with_authority(
                 &id(),
                 &self.stake_pool.pubkey(),
-                &self.validator_list.pubkey(),
-                &self.deposit_authority,
+                &sol_deposit_authority.pubkey(),
                 &self.withdraw_authority,
-                stake,
-                &current_staker.pubkey(),
-                validator_stake_account,
                 &self.reserve_stake.pubkey(),
+                &payer.pubkey(),
                 pool_account,
+                &self.pool_fee_account.pubkey(),
+                &self.pool_fee_account.pubkey(),
                 &self.pool_mint.pubkey(),
                 &spl_token::id(),
+                amount,
             )
         } else {
-            instruction::deposit(
+            instruction::deposit_sol(
                 &id(),
                 &self.stake_pool.pubkey(),
-                &self.validator_list.pubkey(),
                 &self.withdraw_authority,
-                stake,
-                &current_staker.pubkey(),
-                validator_stake_account,
                 &self.reserve_stake.pubkey(),
+                &payer.pubkey(),
                 pool_account,
+                &self.pool_fee_account.pubkey(),
+                &self.pool_fee_account.pubkey(),
                 &self.pool_mint.pubkey(),
                 &spl_token::id(),
+                amount,
             )
         };
         let transaction = Transaction::new_signed_with_payer(
@@ -665,7 +910,7 @@ impl StakePoolAccounts {
         amount: u64,
     ) -> Option<TransportError> {
         let transaction = Transaction::new_signed_with_payer(
-            &[instruction::withdraw(
+            &[instruction::withdraw_stake(
                 &id(),
                 &self.stake_pool.pubkey(),
                 &self.validator_list.pubkey(),
@@ -675,6 +920,7 @@ impl StakePoolAccounts {
                 recipient_new_authority,
                 &user_transfer_authority.pubkey(),
                 pool_account,
+                &self.pool_fee_account.pubkey(),
                 &self.pool_mint.pubkey(),
                 &spl_token::id(),
                 amount,
@@ -686,6 +932,11 @@ impl StakePoolAccounts {
         banks_client.process_transaction(transaction).await.err()
     }
 
+    pub async fn get_validator_list(&self, banks_client: &mut BanksClient) -> ValidatorList {
+        let validator_list_account = get_account(banks_client, &self.validator_list.pubkey()).await;
+        try_from_slice_unchecked::<ValidatorList>(validator_list_account.data.as_slice()).unwrap()
+    }
+
     pub async fn update_validator_list_balance(
         &self,
         banks_client: &mut BanksClient,
@@ -694,6 +945,7 @@ impl StakePoolAccounts {
         validator_vote_accounts: &[Pubkey],
         no_merge: bool,
     ) -> Option<TransportError> {
+        let validator_list = self.get_validator_list(banks_client).await;
         let transaction = Transaction::new_signed_with_payer(
             &[instruction::update_validator_list_balance(
                 &id(),
@@ -701,6 +953,7 @@ impl StakePoolAccounts {
                 &self.withdraw_authority,
                 &self.validator_list.pubkey(),
                 &self.reserve_stake.pubkey(),
+                &validator_list,
                 validator_vote_accounts,
                 0,
                 no_merge,
@@ -727,6 +980,7 @@ impl StakePoolAccounts {
                 &self.reserve_stake.pubkey(),
                 &self.pool_fee_account.pubkey(),
                 &self.pool_mint.pubkey(),
+                &spl_token::id(),
             )],
             Some(&payer.pubkey()),
             &[payer],
@@ -762,6 +1016,7 @@ impl StakePoolAccounts {
         validator_vote_accounts: &[Pubkey],
         no_merge: bool,
     ) -> Option<TransportError> {
+        let validator_list = self.get_validator_list(banks_client).await;
         let transaction = Transaction::new_signed_with_payer(
             &[
                 instruction::update_validator_list_balance(
@@ -770,6 +1025,7 @@ impl StakePoolAccounts {
                     &self.withdraw_authority,
                     &self.validator_list.pubkey(),
                     &self.reserve_stake.pubkey(),
+                    &validator_list,
                     validator_vote_accounts,
                     0,
                     no_merge,
@@ -782,6 +1038,7 @@ impl StakePoolAccounts {
                     &self.reserve_stake.pubkey(),
                     &self.pool_fee_account.pubkey(),
                     &self.pool_mint.pubkey(),
+                    &spl_token::id(),
                 ),
                 instruction::cleanup_removed_validator_entries(
                     &id(),
@@ -802,15 +1059,18 @@ impl StakePoolAccounts {
         payer: &Keypair,
         recent_blockhash: &Hash,
         stake: &Pubkey,
+        validator: &Pubkey,
     ) -> Option<TransportError> {
         let transaction = Transaction::new_signed_with_payer(
             &[instruction::add_validator_to_pool(
                 &id(),
                 &self.stake_pool.pubkey(),
                 &self.staker.pubkey(),
+                &payer.pubkey(),
                 &self.withdraw_authority,
                 &self.validator_list.pubkey(),
                 stake,
+                validator,
             )],
             Some(&payer.pubkey()),
             &[payer, &self.staker],
@@ -827,20 +1087,31 @@ impl StakePoolAccounts {
         new_authority: &Pubkey,
         validator_stake: &Pubkey,
         transient_stake: &Pubkey,
+        destination_stake: &Keypair,
     ) -> Option<TransportError> {
         let transaction = Transaction::new_signed_with_payer(
-            &[instruction::remove_validator_from_pool(
-                &id(),
-                &self.stake_pool.pubkey(),
-                &self.staker.pubkey(),
-                &self.withdraw_authority,
-                &new_authority,
-                &self.validator_list.pubkey(),
-                validator_stake,
-                transient_stake,
-            )],
+            &[
+                system_instruction::create_account(
+                    &payer.pubkey(),
+                    &destination_stake.pubkey(),
+                    0,
+                    std::mem::size_of::<stake_program::StakeState>() as u64,
+                    &stake_program::id(),
+                ),
+                instruction::remove_validator_from_pool(
+                    &id(),
+                    &self.stake_pool.pubkey(),
+                    &self.staker.pubkey(),
+                    &self.withdraw_authority,
+                    &new_authority,
+                    &self.validator_list.pubkey(),
+                    validator_stake,
+                    transient_stake,
+                    &destination_stake.pubkey(),
+                ),
+            ],
             Some(&payer.pubkey()),
-            &[payer, &self.staker],
+            &[payer, &self.staker, destination_stake],
             *recent_blockhash,
         );
         banks_client.process_transaction(transaction).await.err()
@@ -854,6 +1125,7 @@ impl StakePoolAccounts {
         validator_stake: &Pubkey,
         transient_stake: &Pubkey,
         lamports: u64,
+        transient_stake_seed: u64,
     ) -> Option<TransportError> {
         let transaction = Transaction::new_signed_with_payer(
             &[instruction::decrease_validator_stake(
@@ -865,6 +1137,7 @@ impl StakePoolAccounts {
                 validator_stake,
                 transient_stake,
                 lamports,
+                transient_stake_seed,
             )],
             Some(&payer.pubkey()),
             &[payer, &self.staker],
@@ -881,6 +1154,7 @@ impl StakePoolAccounts {
         transient_stake: &Pubkey,
         validator: &Pubkey,
         lamports: u64,
+        transient_stake_seed: u64,
     ) -> Option<TransportError> {
         let transaction = Transaction::new_signed_with_payer(
             &[instruction::increase_validator_stake(
@@ -893,6 +1167,7 @@ impl StakePoolAccounts {
                 transient_stake,
                 validator,
                 lamports,
+                transient_stake_seed,
             )],
             Some(&payer.pubkey()),
             &[payer, &self.staker],
@@ -932,15 +1207,19 @@ pub async fn simple_add_validator_to_pool(
     recent_blockhash: &Hash,
     stake_pool_accounts: &StakePoolAccounts,
 ) -> ValidatorStakeAccount {
-    let validator_stake = ValidatorStakeAccount::new(&stake_pool_accounts.stake_pool.pubkey());
-    validator_stake
-        .create_and_delegate(
-            banks_client,
-            &payer,
-            &recent_blockhash,
-            &stake_pool_accounts.staker,
-        )
-        .await;
+    let validator_stake = ValidatorStakeAccount::new(
+        &stake_pool_accounts.stake_pool.pubkey(),
+        DEFAULT_TRANSIENT_STAKE_SEED,
+    );
+
+    create_vote(
+        banks_client,
+        payer,
+        recent_blockhash,
+        &validator_stake.validator,
+        &validator_stake.vote,
+    )
+    .await;
 
     let error = stake_pool_accounts
         .add_validator_to_pool(
@@ -948,6 +1227,7 @@ pub async fn simple_add_validator_to_pool(
             &payer,
             &recent_blockhash,
             &validator_stake.stake_account,
+            &validator_stake.vote.pubkey(),
         )
         .await;
     assert!(error.is_none());
@@ -1018,8 +1298,8 @@ impl DepositStakeAccount {
         .await;
     }
 
-    pub async fn deposit(
-        &self,
+    pub async fn deposit_stake(
+        &mut self,
         banks_client: &mut BanksClient,
         payer: &Keypair,
         recent_blockhash: &Hash,
@@ -1048,11 +1328,12 @@ impl DepositStakeAccount {
                 &self.authority,
             )
             .await;
+        self.pool_tokens = get_token_balance(banks_client, &self.pool_account.pubkey()).await;
         assert!(error.is_none());
     }
 }
 
-pub async fn simple_deposit(
+pub async fn simple_deposit_stake(
     banks_client: &mut BanksClient,
     payer: &Keypair,
     recent_blockhash: &Hash,
